import pytest
import pandas as pd

from helper import check_col_values, setup, check_base_cols_present, Presets

setup()

import markets_insights
from markets_insights.core.column_definition import (
    AggregationPeriods,
    BaseColumns,
    CalculatedColumns,
)
from markets_insights.datareader.data_reader import NseIndicesReader
from markets_insights.dataprocess.data_processor import (
    HistoricalDataProcessOptions,
    HistoricalDataProcessor,
    HistoricalDataset,
)


def test_historical_data_processor_without_options():
    processor = HistoricalDataProcessor()
    result: HistoricalDataset = processor.process(
        NseIndicesReader(),
        {"from_date": Presets.dates.dec_start, "to_date": Presets.dates.dec_end},
    )
    check_base_cols_present(result.get_daily_data(), "Daily")
    assert result.get_daily_data().shape[0] == 2146

    check_base_cols_present(result.get_monthly_data(), AggregationPeriods.Monthly)
    assert result.get_monthly_data().shape[0] == 108

    check_base_cols_present(result.get_annual_data(), AggregationPeriods.Annual)
    assert result.get_monthly_data().shape[0] == 108


@pytest.mark.parametrize(
    "options",
    [
        HistoricalDataProcessOptions(include_monthly_data=False),
        HistoricalDataProcessOptions(include_annual_data=False),
        HistoricalDataProcessOptions(
            include_monthly_data=False, include_annual_data=False
        ),
    ],
)
def test_historical_data_processor_with_options(options: HistoricalDataProcessOptions):
    processor = HistoricalDataProcessor(options)
    result = processor.process(
        NseIndicesReader(),
        {"from_date": Presets.dates.dec_start, "to_date": Presets.dates.dec_end},
    )
    check_base_cols_present(result.get_daily_data(), "Daily")

    if options.include_monthly_data:
        check_base_cols_present(result.get_monthly_data(), AggregationPeriods.Monthly)
    else:
        assert result.get_monthly_data() is None

    if options.include_annual_data:
        check_base_cols_present(result.get_annual_data(), AggregationPeriods.Annual)
    else:
        assert result.get_annual_data() is None


def test_historical_data_processor_monthly_aggregration():
    processor = HistoricalDataProcessor()
    result = processor.process(
        NseIndicesReader(),
        {"from_date": Presets.dates.dec_start, "to_date": Presets.dates.dec_end},
    )

    check_col_values(
        data=result.get_monthly_data().query(
            str(Presets.filters.nifty50 & Presets.filters.date_year_end)
        ),
        col_value_pairs={
            BaseColumns.Open: 20194.1,
            BaseColumns.High: 21801.45,
            BaseColumns.Low: 20183.7,
            BaseColumns.Close: 21731.4,
            BaseColumns.Turnover: 5752296300000.0,
        },
    )


def test_historical_data_processor_annual_aggregration():
    processor = HistoricalDataProcessor()
    result = processor.process(
        NseIndicesReader(),
        {"from_date": Presets.dates.year_start, "to_date": Presets.dates.year_end},
    )

    check_col_values(
        data=result.get_annual_data().query(
            str(Presets.filters.nifty50 & Presets.filters.date_year_end)
        ),
        col_value_pairs={
            BaseColumns.Open: 18131.7,
            BaseColumns.High: 21801.45,
            BaseColumns.Low: 16828.35,
            BaseColumns.Close: 21731.4,
            BaseColumns.Turnover: 58075646100000.0,
        },
    )



from datetime import date
import os
import sys


code_dir = os.path.join(os.path.abspath("."), "src")
os.chdir(code_dir)
sys.path.append(code_dir)

from markets_insights.dataprocess.data_processor import HistoricalDataProcessor
from markets_insights.datareader.data_reader import NseIndicesReader

processor = HistoricalDataProcessor()
result = processor.process(
    NseIndicesReader(), {"from_date": date(2023, 1, 1), "to_date": date(2023, 12, 31)}
)



from string import Template


class AggregationPeriods:
    Monthly = "Monthly"
    Annual = "Annual"


class BasePriceColumns:
    Open = "Open"
    High = "High"
    Low = "Low"
    Close = "Close"


class BaseColumns(BasePriceColumns):
    Identifier = "Identifier"
    PreviousClose = "PreviousClose"
    Date = "Date"
    Turnover: str = "Turnover"
    Volume: str = "Volume"


PeriodAggregateColumnTemplate: Template = Template("$period $col_name")


class CalculatedColumnsBase:
    Month = "Month"
    MonthNo = "MonthNo"
    Year = "Year"
    Day = "Day"
    Vwap = "Vwap"
    PriceCrossedAboveVwap = "PriceCrossedAboveVwap"


class CalculatedColumns(CalculatedColumnsBase):
    Change = "Change"
    IsInDerivatives = "IsInDerivatives"
    ClosePriceDiff = "ClosePriceDiff"
    Gain = "Gain"
    Loss = "Loss"
    AvgGain = "AvgGain"
    AvgLoss = "AvgLoss"
    RelativeStrength = "Rs"
    RelativeStrengthIndex = "Rsi"
    RsiCrossedAbove = "RsiCrossedAbove"
    RsiCrossedBelow = "RsiCrossedBelow"
    StochRsi_K = "StochRsi_K"
    StochRsi_D = "StochRsi_D"
    StochRsi_KCrossedAbove = "StochRsi_KCrossedAbove"
    StochRsi_KCrossedBelow = "StochRsi_KCrossedBelow"
    StochRsi_DCrossedAbove = "StochRsi_DCrossedAbove"
    StochRsi_DCrossedBelow = "StochRsi_DCrossedBelow"


class DerivativesCalculatedColumns(CalculatedColumnsBase):
    CloseToPrevCloseChangePerc: str = "CloseToPrevCloseChangePerc"
    OpenToPrevCloseChangePerc: str = "OpenToPrevCloseChangePerc"
    PreviousCloseAmount: str = "PreviousCloseAmount"
    OpenAmount: str = "OpenAmount"
    CloseAmount: str = "CloseAmount"
    HighAmount: str = "HighAmount"
    LowAmount: str = "LowAmount"
    AmountDiffOpenToPrevClose: str = "AmountDiffOpenToPrevClose"
    AmountDiffCloseToPrevClose: str = "AmountDiffCloseToPrevClose"
    AmountDiffOpenToPrevClose: str = "AmountDiffOpenToPrevClose"
    ClosestStrike: str = "ClosestStrike"
    LotSize: str = "LotSize"


class DerivativesBaseColumns(BaseColumns):
    InstrumentType: str = "InstrumentType"
    AveragePrice: str = "AvrgPric"
    OptionType: str = "OptionType"
    ExpiryDate: str = "ExpiryDate"
    StrikePrice: str = "StrkPric"
    OpenInterest: str = "OpenInterest"
    OiChangePct: float = "OiChangePerc"



from datetime import date
import pytest
import pandas as pd

from helper import check_col_values, setup, Presets
from markets_insights.calculations.base import (
    PriceCrossedAboveValueFlagWorker,
    PriceCrossedBelowValueFlagWorker,
    StochRsiCalculationWorker,
    VwapCalculationWorker,
)
from markets_insights.core.core import DateFilter, IdentifierFilter

setup()

import markets_insights
from markets_insights.core.column_definition import CalculatedColumns
from markets_insights.datareader.data_reader import BhavCopyReader, NseIndicesReader
from markets_insights.dataprocess.data_processor import (
    CalculationPipelineBuilder,
    HistoricalDataProcessor,
    HistoricalDataset,
    MultiDataCalculationPipelines,
)

indices_processor = HistoricalDataProcessor()
indices_result: HistoricalDataset = indices_processor.process(
    NseIndicesReader(),
    {"from_date": Presets.dates.year_start, "to_date": Presets.dates.year_end},
)

equity_processor = HistoricalDataProcessor()
equity_result: HistoricalDataset = equity_processor.process(
    BhavCopyReader(),
    {"from_date": Presets.dates.q4_start, "to_date": Presets.dates.q4_end},
)


def test_calculations_rsi():
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    pipelines.set_item(
        "rsi", CalculationPipelineBuilder.create_rsi_calculation_pipeline()
    )
    indices_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    indices_processor.run_calculation_pipelines()
    daily_data = indices_result.get_daily_data()

    check_col_values(
        data=daily_data.query(
            str(Presets.filters.nifty50 & Presets.filters.date_year_end)
        ),
        col_value_pairs={CalculatedColumns.RelativeStrengthIndex: 76.39132974363618},
    )


def test_calculations_crossing_flags():
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    pipelines.set_item(
        "rsi", CalculationPipelineBuilder.create_rsi_calculation_pipeline()
    )
    indices_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    indices_processor.run_calculation_pipelines()
    daily_data = indices_result.get_daily_data()

    for for_date in [date(2023, 2, 28), date(2023, 10, 26)]:
        check_col_values(
            data=daily_data.query(
                str(Presets.filters.nifty50 & DateFilter(for_date=for_date))
            ),
            col_value_pairs={"RsiCrossedBelow": True},
        )

    for for_date in [
        date(2023, 7, 4),
        date(2023, 9, 15),
        date(2023, 12, 1),
        date(2023, 12, 27),
    ]:
        check_col_values(
            data=daily_data.query(
                str(Presets.filters.nifty50 & DateFilter(for_date=for_date))
            ),
            col_value_pairs={"RsiCrossedAbove": True},
        )


def test_calculations_sma():
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    pipelines.set_item(
        "sma", CalculationPipelineBuilder.create_sma_calculation_pipeline()
    )
    indices_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    indices_processor.run_calculation_pipelines()
    daily_data = indices_result.get_daily_data()

    check_col_values(
        data=daily_data.query(
            str(Presets.filters.nifty50 & Presets.filters.date_year_end)
        ),
        col_value_pairs={
            "Sma50": 20171.113,
            "Sma100": 19891.132,
            "Sma200": 19142.190749999998,
        },
    )


def test_calculations_stoch_rsi():
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    stoch_rsi_worker = StochRsiCalculationWorker()
    pipelines.set_item(
        "stoch_rsi",
        CalculationPipelineBuilder.create_pipeline_for_worker(stoch_rsi_worker),
    )
    indices_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    indices_processor.run_calculation_pipelines()
    daily_data = indices_result.get_daily_data()

    check_col_values(
        data=daily_data.query(
            str(Presets.filters.nifty50 & Presets.filters.date_year_end)
        ),
        col_value_pairs={
            CalculatedColumns.StochRsi_K: 53.10131491044907,
            CalculatedColumns.StochRsi_D: 45.248290876233135,
        },
    )


@pytest.mark.parametrize(
    "identifier,for_date,vwap",
    [
        ("RELIANCE", date(2023, 12, 21), 2383.908377),
        ("INFY", date(2023, 12, 27), 1457.663944),
        ("HDFCBANK", date(2023, 12, 29), 1572.029516),
    ],
)
def test_calculations_vwap(identifier: str, for_date: date, vwap: float):
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    worker = VwapCalculationWorker(time_window=50)
    pipelines.set_item(
        "vwap", CalculationPipelineBuilder.create_pipeline_for_worker(worker)
    )
    equity_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    equity_processor.run_calculation_pipelines()
    daily_data = equity_result.get_daily_data()

    check_col_values(
        data=daily_data.query(str(IdentifierFilter(identifier) & DateFilter(for_date))),
        col_value_pairs={CalculatedColumns.Vwap: vwap},
    )


def test_calculations_bb():
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    pipelines.set_item(
        "bb", CalculationPipelineBuilder.create_bb_calculation_pipeline()
    )
    indices_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    indices_processor.run_calculation_pipelines()
    daily_data = indices_result.get_daily_data()

    check_col_values(
        data=daily_data.query(
            str(Presets.filters.nifty50 & Presets.filters.date_year_end)
        ),
        col_value_pairs={
            "StdDev200": 1084.4860011940468,
            "Bb200Dev2Upper": 21311.162752388092,
            "Bb200Dev2Lower": 16973.218747611903,
            "Bb200Dev3Upper": 22395.64875358214,
            "Bb200Dev3Lower": 15888.732746417858,
        },
    )


@pytest.mark.parametrize(
    "n_day,lowest_price,price_fall_pert",
    [
        (5, 20507.75, -1.183398378717077),
        (10, 20507.75, -1.183398378717077),
        (15, 20507.75, -1.183398378717077),
    ],
)
def test_calculations_fwd_looking_fall(
    n_day: int, lowest_price: float, price_fall_pert: float
):
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    pipelines.set_item(
        "fwd_price_fall",
        CalculationPipelineBuilder.create_forward_looking_price_fall_pipeline([n_day]),
    )
    indices_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    indices_processor.run_calculation_pipelines()
    daily_data = indices_result.get_daily_data()

    test_day_data = daily_data.query(
        str(Presets.filters.nifty50 & DateFilter(for_date=Presets.dates.dec_start))
    )
    check_col_values(
        data=test_day_data,
        col_value_pairs={
            f"TroughInNext{str(n_day)}Sessions": lowest_price,
            f"TroughPercInNext{str(n_day)}Sessions": price_fall_pert,
        },
    )


@pytest.mark.parametrize(
    "n_day,highest_price,price_rise_pert",
    [
        (5, 21006.1, 3.6422125627223196),
        (10, 21492.3, 6.041079736923893),
        (15, 21593.0, 6.537924501305012),
    ],
)
def test_calculations_fwd_looking_rise(
    n_day: int, highest_price: float, price_rise_pert: float
):
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    pipelines.set_item(
        "fwd_price_rise",
        CalculationPipelineBuilder.create_forward_looking_price_rise_pipeline([n_day]),
    )
    indices_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    indices_processor.run_calculation_pipelines()
    daily_data = indices_result.get_daily_data()

    test_day_data = daily_data.query(
        str(Presets.filters.nifty50 & DateFilter(for_date=Presets.dates.dec_start))
    )
    check_col_values(
        data=test_day_data,
        col_value_pairs={
            f"PeakInNext{str(n_day)}Sessions": highest_price,
            f"PeakPercInNext{str(n_day)}Sessions": price_rise_pert,
        },
    )


@pytest.mark.parametrize(
    "identifier,for_date,flag",
    [
        ("WONDERLA", date(2023, 12, 14), True),
        ("GAEL", date(2023, 12, 28), True),
        ("RELIANCE", date(2023, 12, 14), False),
        ("INFY", date(2023, 12, 29), False),
    ],
)
def test_calculations_price_crossing_below_flag(
    identifier: str, for_date: date, flag: bool
):
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    pipelines.set_item(
        "vwap",
        CalculationPipelineBuilder.create_pipeline_for_worker(
            VwapCalculationWorker(time_window=50)
        ),
    )
    pipelines.set_item(
        "vwap_crossed",
        CalculationPipelineBuilder.create_pipeline_for_worker(
            PriceCrossedBelowValueFlagWorker(value_column=CalculatedColumns.Vwap)
        ),
    )
    equity_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    equity_processor.run_calculation_pipelines()
    daily_data = equity_result.get_daily_data()

    check_col_values(
        data=daily_data.query(
            str(IdentifierFilter(identifier) & DateFilter(for_date=for_date))
        ),
        col_value_pairs={"PriceCrossedBelowVwap": flag},
    )


@pytest.mark.parametrize(
    "identifier,for_date,flag",
    [
        ("ZYDUSWELL", date(2023, 12, 14), True),
        ("HINDZINC", date(2023, 12, 29), True),
        ("RELIANCE", date(2023, 12, 14), False),
        ("INFY", date(2023, 12, 29), False),
    ],
)
def test_calculations_price_crossing_above_flag(
    identifier: str, for_date: date, flag: bool
):
    # Prepare calculation pipeline
    pipelines = MultiDataCalculationPipelines()
    pipelines.set_item(
        "vwap",
        CalculationPipelineBuilder.create_pipeline_for_worker(
            VwapCalculationWorker(time_window=50)
        ),
    )
    pipelines.set_item(
        "vwap_crossed",
        CalculationPipelineBuilder.create_pipeline_for_worker(
            PriceCrossedAboveValueFlagWorker(value_column=CalculatedColumns.Vwap)
        ),
    )
    equity_processor.set_calculation_pipelines(pipelines)

    # Run the pipeline and get data
    equity_processor.run_calculation_pipelines()
    daily_data = equity_result.get_daily_data()

    check_col_values(
        data=daily_data.query(
            str(IdentifierFilter(identifier) & DateFilter(for_date=for_date))
        ),
        col_value_pairs={"PriceCrossedAboveVwap": flag},
    )



from markets_insights.core.core import Instrumentation
from markets_insights.core.column_definition import BaseColumns, CalculatedColumns
import pandas as pd
import pandas_ta as ta


class CalculationWorker:
    def __init__(self, **params):
        self._columns: [str] = []
        self._params: dict = params

    def get_columns(self) -> [str]:
        return self._columns
    
    def get_column(self) -> str:
        return self._columns[0]
    
    def get_params(self) -> dict:
        return self._params

    def add_calculated_columns(self, data: pd.DataFrame):
        raise NotImplementedError("add_calculated_fields")


class CalculationPipeline:
    _pipeline: list

    def __init__(self, workers=None):
        if workers is not None:
            self._pipeline = workers
        else:
            self._pipeline = []

    def add_calculation_worker(self, worker: CalculationWorker):
        self._pipeline.append(worker)

    def run(self, data):
        for worker in self._pipeline:
            result = worker.add_calculated_columns(data)
            if result is not None:
                data = result
        return data


class ValueCrossedAboveFlagWorker(CalculationWorker):
    def __init__(self, value_column: str, value):
        super().__init__(value_column = value_column, value = value)
        self._columns.append(f"{value_column}CrossedAbove")

    @Instrumentation.trace(name="ValueCrossedAboveFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(BaseColumns.Identifier)
        data[self._columns[0]] = identifier_grouped_data[self._params['value_column']].transform(
            lambda x: (x.shift(1) < self._params['value']) & (x >= self._params['value'])
        )


class ValueCrossedBelowFlagWorker(CalculationWorker):
    def __init__(self, value_column: str, value):
        super().__init__(value_column = value_column, value = value)
        self._columns.append(f"{value_column}CrossedBelow")

    @Instrumentation.trace(name="ValueCrossedBelowFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(BaseColumns.Identifier)
        data[self._columns[0]] = identifier_grouped_data[self._params['value_column']].transform(
            lambda x: (x.shift(-1) > self._params['value']) & (x <= self._params['value'])
        )


class PriceCrossedAboveValueFlagWorker(CalculationWorker):
    def __init__(self, value_column: str):
        super().__init__(value_column = value_column)
        self._columns.append(f"PriceCrossedAbove{value_column}")

    @Instrumentation.trace(name="PriceCrossedAboveValueFlagWorker")
    def add_calculated_columns(self, data):
        data[self._columns[0]] = (
            data[BaseColumns.Close] >= data[self._params['value_column']]
        ) & (data[BaseColumns.PreviousClose] < data[self._params['value_column']])


class PriceCrossedBelowValueFlagWorker(CalculationWorker):
    def __init__(self, value_column: str):
        super().__init__(value_column = value_column)
        self._columns.append(f"PriceCrossedBelow{value_column}")

    @Instrumentation.trace(name="PriceCrossedBelowValueFlagWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[self._columns[0]] = (
            data[BaseColumns.Close] < data[self._params['value_column']]
        ) & (data[BaseColumns.PreviousClose] >= data[self._params['value_column']])


class DatePartsCalculationWorker(CalculationWorker):
    def __init__(self):
        super().__init__()
        self._columns.append(CalculatedColumns.Year)
        self._columns.append(CalculatedColumns.MonthNo)
        self._columns.append(CalculatedColumns.Month)
        self._columns.append(CalculatedColumns.Day)

    @Instrumentation.trace(name="DatePartsCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[CalculatedColumns.Year] = data[BaseColumns.Date].dt.year
        data[CalculatedColumns.MonthNo] = data[BaseColumns.Date].dt.month
        data[CalculatedColumns.Month] = data[BaseColumns.Date].dt.strftime("%b")
        data[CalculatedColumns.Day] = data[BaseColumns.Date].dt.strftime("%A")


class SmaCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 50):
        super().__init__(time_window = time_window)
        self._columns.append(f"Sma{str(time_window)}")

    @Instrumentation.trace(name="SmaCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(BaseColumns.Identifier)
        data[self._columns[0]] = identifier_grouped_data[BaseColumns.Close].transform(
            lambda x: x.rolling(self._params['time_window']).mean()
        )


class StdDevCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 200):
        super().__init__(time_window = time_window)
        self._columns.append(f"StdDev{time_window}")

    @Instrumentation.trace(name="StdDevCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        identifier_grouped_data = data.groupby(BaseColumns.Identifier)
        data[self._columns[0]] = identifier_grouped_data[BaseColumns.Close].transform(
            lambda x: x.rolling(self._params['time_window']).std()
        )


class BollingerBandCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 200, deviation: int = 2):
        super().__init__(time_window = time_window, deviation = deviation)
        self._columns.append(f"Bb{str(time_window)}Dev{str(deviation)}Lower")
        self._columns.append(f"Bb{str(time_window)}Dev{str(deviation)}Upper")

    @Instrumentation.trace(name="BollingerBandCalculationWorker")
    def add_calculated_columns(self, data: pd.DataFrame):
        data[self._columns[0]] = data[f"Sma{str(self._params['time_window'])}"] - (
            data[f"StdDev{str(self._params['time_window'])}"] * self._params['deviation']
        )
        data[self._columns[1]] = data[f"Sma{str(self._params['time_window'])}"] + (
            data[f"StdDev{str(self._params['time_window'])}"] * self._params['deviation']
        )


class RsiOldCalculationWorker(CalculationWorker):
    def __init__(self, time_window):
        super().__init__()
        self._time_window = time_window
        self._column_name = "Rsi" + str(self._time_window)

    def calculate_wsm_average(self, raw_data, data, avg_col_name, abs_col_name):
        step = 1
        for i, row in enumerate(data[avg_col_name].iloc[self._time_window + step :]):
            raw_data.at[data.index[i + self._time_window + step], avg_col_name] = (
                data[avg_col_name].iloc[i + self._time_window]
                * (self._time_window - step)
                + data[abs_col_name].iloc[i + self._time_window + step]
            ) / self._time_window

    @Instrumentation.trace(name="RsiOldCalculationWorker")
    def add_calculated_columns(self, data):
        identifier_grouped_data = data.groupby(BaseColumns.Identifier)
        data[CalculatedColumns.ClosePriceDiff] = identifier_grouped_data[
            BaseColumns.Close
        ].transform(lambda x: x.diff(1))
        data[CalculatedColumns.Gain] = identifier_grouped_data[
            CalculatedColumns.ClosePriceDiff
        ].transform(lambda x: x.clip(lower=0).round(2))
        data[CalculatedColumns.Loss] = identifier_grouped_data[
            CalculatedColumns.ClosePriceDiff
        ].transform(lambda x: x.clip(upper=0).abs().round(2))

        data[CalculatedColumns.ClosePriceDiff].fillna(0)
        data[CalculatedColumns.Gain].fillna(0)
        data[CalculatedColumns.Loss].fillna(0)

        # Get initial Averages
        data[CalculatedColumns.AvgGain] = identifier_grouped_data[
            CalculatedColumns.Gain
        ].transform(
            lambda x: x.rolling(
                window=self._time_window, min_periods=self._time_window
            ).mean()
        )
        data[CalculatedColumns.AvgLoss] = identifier_grouped_data[
            CalculatedColumns.Loss
        ].transform(
            lambda x: x.rolling(
                window=self._time_window, min_periods=self._time_window
            ).mean()
        )

        for identifier in data[BaseColumns.Identifier].unique():
            self.calculate_wsm_average(
                data,
                identifier_grouped_data.get_group(identifier),
                CalculatedColumns.AvgGain,
                CalculatedColumns.Gain,
            )
            self.calculate_wsm_average(
                data,
                identifier_grouped_data.get_group(identifier),
                CalculatedColumns.AvgLoss,
                CalculatedColumns.Loss,
            )

        data[CalculatedColumns.RelativeStrength] = (
            data[CalculatedColumns.AvgGain] / data[CalculatedColumns.AvgLoss]
        )
        data[CalculatedColumns.RelativeStrengthIndex] = 100 - (
            100 / (1.0 + data[CalculatedColumns.RelativeStrength])
        )


class RsiCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 14):
        super().__init__(time_window = time_window)
        self._columns.append(CalculatedColumns.RelativeStrengthIndex)

    def calculate_rsi(self, group):
        group[CalculatedColumns.RelativeStrengthIndex] = ta.rsi(
            group[BaseColumns.Close]
        )
        return group

    @Instrumentation.trace(name="RsiCalculationWorker")
    def add_calculated_columns(self, data):
        result = data.groupby([BaseColumns.Identifier], group_keys=True).apply(
            self.calculate_rsi
        )
        return result.reset_index(drop=True)


class StochRsiCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 14):
        super().__init__(time_window = time_window)
        self._time_window = time_window
        self._columns.append(CalculatedColumns.StochRsi_K)
        self._columns.append(CalculatedColumns.StochRsi_D)

    def calculate_stoch_rsi(self, group: pd.DataFrame):
        data = ta.stochrsi(group[BaseColumns.Close], window=14, smooth1=3, smooth2=3)
        if data is not None:
            group[CalculatedColumns.StochRsi_K] = data["STOCHRSIk_14_14_3_3"]
            group[CalculatedColumns.StochRsi_D] = data["STOCHRSId_14_14_3_3"]
        return group

    @Instrumentation.trace(name="StochRsiCalculationWorker")
    def add_calculated_columns(self, data):
        result = data.groupby([BaseColumns.Identifier], group_keys=True).apply(
            self.calculate_stoch_rsi
        )
        return result.reset_index(drop=True)


class VwapCalculationWorker(CalculationWorker):
    def __init__(self, time_window: int = 200):
        super().__init__(time_window = time_window)
        self._time_window = time_window
        self._columns.append(CalculatedColumns.Vwap)

    @Instrumentation.trace(name="VwapCalculationWorker")
    def add_calculated_columns(self, data):
        data[BaseColumns.Turnover] = data[BaseColumns.Turnover].replace("-", 0)
        data[BaseColumns.Volume] = data[BaseColumns.Volume].replace("-", 0)
        data[CalculatedColumns.Vwap] = (
            data.groupby(BaseColumns.Identifier)
            .apply(
                lambda x: x[BaseColumns.Turnover].rolling(self._time_window).sum()
                / x.rolling(self._time_window)[BaseColumns.Volume].sum()
            )
            .reset_index(level=0, drop=True)
        )


class LowestPriceInNextNDaysCalculationWorker(CalculationWorker):
    def __init__(self, time_window):
        super().__init__(time_window = time_window)
        self._columns.append(f"TroughInNext{str(time_window)}Sessions")
        self._columns.append(f"TroughPercInNext{str(time_window)}Sessions")

    @Instrumentation.trace(name="LowestPriceInNextNDaysCalculationWorker")
    def add_calculated_columns(self, data):
        identifier_grouped_data = data.groupby(BaseColumns.Identifier)
        data[self._columns[0]] = identifier_grouped_data[
            BaseColumns.Low
        ].transform(
            lambda x: x.rolling(self._params['time_window']).min().shift(-self._params['time_window'])
        )
        data[self._columns[1]] = (
            (data[BaseColumns.Close] - data[self._columns[0]])
            / data[BaseColumns.Close]
            * 100
        )


class HighestPriceInNextNDaysCalculationWorker(CalculationWorker):
    def __init__(self, time_window):
        super().__init__(time_window = time_window)
        self._columns.append(f"PeakInNext{str(time_window)}Sessions")
        self._columns.append(f"PeakPercInNext{str(time_window)}Sessions")

    @Instrumentation.trace(name="HighestPriceInNextNDaysCalculationWorker")
    def add_calculated_columns(self, data):
        identifier_grouped_data = data.groupby(BaseColumns.Identifier)
        data[self._columns[0]] = identifier_grouped_data[
            BaseColumns.High
        ].transform(
            lambda x: x.rolling(self._params['time_window']).max().shift(-self._params['time_window'])
        )
        data[self._columns[1]] = (
            (data[self._columns[0]] - data[BaseColumns.Close])
            / data[BaseColumns.Close]
            * 100
        )



import os
import sys

code_dir = os.path.join(os.path.abspath("."), "src")
os.chdir(code_dir)
sys.path.append(code_dir)

import datetime

import markets_insights

from markets_insights.datareader.data_reader import DateRangeDataReader
from markets_insights.core.core import Instrumentation
from markets_insights.core.column_definition import BaseColumns

import yfinance as yf
import pandas


class NasdaqDataReader(DateRangeDataReader):
    def __init__(
        self, tickers: list = ["AAPL", "GOOGL", "MSFT", "AMZN", "META", "TSLA", "NVDA"]
    ):
        super().__init__(reader=None)
        self.tickers = tickers
        self.name = "NasdaqDataReader"

    @Instrumentation.trace(name="NasdaqDataReader.read")
    def read(self, from_date, to_date):
        df_list = list()
        for ticker in self.tickers:
            data = yf.download(ticker, group_by="Ticker", start=from_date, end=to_date)
            data["ticker"] = ticker
            df_list.append(data)

        # combine all dataframes into a single dataframe
        df = pandas.concat(df_list)

        data = df.reset_index().rename(columns=self.get_column_name_mappings())
        data[BaseColumns.Date] = pandas.to_datetime(data[BaseColumns.Date])
        return data

    def get_column_name_mappings(self):
        return {
            "ticker": BaseColumns.Identifier,
            "OPEN": BaseColumns.Open,
            "HIGH": BaseColumns.High,
            "LOW": BaseColumns.Low,
            "CLOSE": BaseColumns.Close,
        }


# import classes & setup options
from markets_insights.dataprocess.data_processor import (
    HistoricalDataProcessor,
    MultiDataCalculationPipelines,
    CalculationPipelineBuilder,
    HistoricalDataProcessOptions,
)
from markets_insights.calculations.base import DatePartsCalculationWorker

reader = NasdaqDataReader()
options = HistoricalDataProcessOptions()
options.include_monthly_data = False
options.include_annual_data = False
histDataProcessor = HistoricalDataProcessor(options)

# Fetch the data
year_start = datetime.date(2023, 1, 1)
to_date = datetime.date(2023, 12, 31)
result = histDataProcessor.process(
    reader, {"from_date": year_start, "to_date": to_date}
)

# Prepare calculation pipeline
pipelines = MultiDataCalculationPipelines()
pipelines.set_item(
    "date_parts",
    CalculationPipelineBuilder.create_pipeline_for_worker(DatePartsCalculationWorker()),
)
pipelines.set_item("rsi", CalculationPipelineBuilder.create_rsi_calculation_pipeline())
histDataProcessor.set_calculation_pipelines(pipelines)

# Run the pipeline
histDataProcessor.run_calculation_pipelines()



from markets_insights.core.column_definition import (
    DerivativesBaseColumns,
    DerivativesCalculatedColumns,
)
from markets_insights.core.core import TypeHelper
import pandas as pd
import datetime
import os


class TradesResultBaseMetrics:
    pnl: float = 0
    capital_required: float = 0
    no_of_trades: int = 0
    non_tradables_no_entry_count: int = 0
    non_tradables_no_exit_count: int = 0
    turnover: float = 0


class TradesResultMetrics(TradesResultBaseMetrics):
    transaction_fees: float = 0
    brokerage: float = 0
    total_expense: float = 0
    net_pnl: float = 0
    roi: float = 0
    break_even_date: datetime.datetime = None


class TradesResultBase(TradesResultMetrics):
    summary: pd.DataFrame

    def __init__(self):
        self.summary = None

    def calculate(self):
        self.transaction_fees = round(self.turnover * 0.01)
        self.brokerage = round(self.no_of_trades * 20)
        self.total_expense = round(self.transaction_fees + self.no_of_trades)
        self.net_pnl = round(self.pnl - self.total_expense)
        if self.capital_required > 0:
            self.roi = round(self.net_pnl / self.capital_required * 100)

    def prepare_summary_df(self):
        self.calculate()
        summary_dict: dict = {}

        props = TypeHelper.get_class_props(TradesResultMetrics)

        for prop in props:
            summary_dict[prop] = getattr(self, prop)

        self.summary = pd.DataFrame(summary_dict, index=[0])

    def __str__(self):
        return f"Net PnL: {round(self.net_pnl)}, Capital Required: {round(self.capital_required)}, ROI: {round(self.roi)}%, Non tradables (No Entry): {self.non_tradables_no_entry_count}, Non tradables (No Exit): {self.non_tradables_no_exit_count}"


class TradesResult(TradesResultBase):
    trades: pd.DataFrame
    non_tradables: pd.DataFrame
    full_trades: pd.DataFrame

    def __init__(self, trades, full_trades, non_tradables):
        super().__init__()
        self.trades = trades
        self.non_tradables = non_tradables
        self.full_trades = full_trades

    def caclulate_from_df(self):
        if len(self.trades) > 0:
            self.trades = self.trades.sort_values(
                DerivativesBaseColumns.Date
            ).reset_index(drop=True)
            self.trades["CumCredit"] = self.trades["Credit"].cumsum()
            self.capital_required = round(abs(self.trades["CumCredit"].min()))
            self.no_of_trades = self.trades[DerivativesBaseColumns.Identifier].count()
            self.pnl = round(self.full_trades["Net"].sum())
            self.turnover = round(self.full_trades["Turnover"].sum())
            self.break_even_date = self.get_break_even_date()

        super().calculate()

    def get_break_even_date(self):
        break_even_date = (
            self.trades[self.trades["CumCredit"] <= 0]
            .tail(1)[DerivativesBaseColumns.Date]
            .values[0]
        )
        return pd.to_datetime(break_even_date)

    def prepare_summary(self):
        self.caclulate_from_df()

        if (len(self.non_tradables)) > 0:
            self.non_tradables_no_entry_count = self.non_tradables[
                self.non_tradables["reason"] == "No Entry"
            ][DerivativesBaseColumns.Identifier].count()
            self.non_tradables_no_exit_count = self.non_tradables[
                self.non_tradables["reason"] == "No Exit"
            ][DerivativesBaseColumns.Identifier].count()

        super().prepare_summary_df()

    def save_result(self, excel_writer: pd.ExcelWriter, suffix: str):
        if self.summary is not None:
            self.summary.to_excel(excel_writer, f"{suffix}-summary")

        self.trades.to_excel(excel_writer, f"{suffix}-trades")
        self.full_trades.to_excel(excel_writer, f"{suffix}-full-trades")
        self.non_tradables.to_excel(excel_writer, f"{suffix}-non-tradables")

    def load_from_file(file_path: str, suffix: str):
        trades = pd.read_excel(file_path, f"{suffix}-trades")
        full_trades = pd.read_excel(file_path, f"{suffix}-full-trades")
        non_tradables = pd.read_excel(file_path, f"{suffix}-non-tradables")
        instance = TradesResult(trades, full_trades, non_tradables)
        instance.prepare_summary()
        return instance


class TradesResultComposite(TradesResult):
    results: dict
    underlying_summaries: pd.DataFrame = None

    def __init__(self):
        super().__init__(pd.DataFrame(), pd.DataFrame(), pd.DataFrame())
        self.results = {}
        self.underlying_summaries = pd.DataFrame()

    def add_result(self, result: TradesResult, key: str):
        self.results[key] = result

        if len(result.trades) > 0:
            result.trades["id"] = key
            self.trades = pd.concat([self.trades, result.trades], ignore_index=True)

        if len(result.non_tradables) > 0:
            result.non_tradables["id"] = key
            self.non_tradables = pd.concat(
                [self.non_tradables, result.non_tradables], ignore_index=True
            )

        if len(result.full_trades) > 0:
            result.full_trades["id"] = key
            self.full_trades = pd.concat(
                [self.full_trades, result.full_trades], ignore_index=True
            )

        if result.summary is not None and len(result.summary) > 0:
            result.summary["id"] = key
            self.underlying_summaries = pd.concat(
                [self.underlying_summaries, result.summary], ignore_index=True
            )

        self.prepare_summary()

    def get_underlying_summaries(self):
        return self.underlying_summaries

    def save_result(self, path: str):
        excel_writer = pd.ExcelWriter(path)
        super().save_result(excel_writer, "")
        self.get_underlying_summaries().to_excel(excel_writer, "underlying-summaries")
        excel_writer.close()


class BiDirectionalTradesResult(TradesResultComposite):
    bear_trades_result: TradesResult
    bull_trades_result: TradesResult

    def __init__(
        self, bear_trades_result: TradesResult, bull_trades_result: TradesResult
    ):
        super().__init__()
        self.bear_trades_result = bear_trades_result
        self.bull_trades_result = bull_trades_result
        self.add_result(bear_trades_result, "bear")
        self.add_result(bull_trades_result, "bull")

    def load_from_file(base_path: str):
        bear_trades_result = TradesResult.load_from_file(base_path, "bear")
        bull_trades_result = TradesResult.load_from_file(base_path, "bull")
        return BiDirectionalTradesResult(bear_trades_result, bull_trades_result)

    def save_result(self, path):
        excel_writer = pd.ExcelWriter(path)
        self.summary.to_excel(excel_writer, f"full-summary")

        for key in self.results:
            result = self.results[key]
            result.save_result(excel_writer, key)

        excel_writer.close()



from markets_insights.core.column_definition import (
    DerivativesBaseColumns,
    DerivativesCalculatedColumns,
)
from markets_insights.core.core import TypeHelper, MarketDaysHelper
import datetime
import pandas as pd
import numpy as np

from markets_insights.trade_builders.results import (
    BiDirectionalTradesResult,
    TradesResultComposite,
    TradesResult,
)


class DerivativeTradeOptions:
    holding_days: int = 7
    budget_min: int = 8000
    budget_max: int = 12000
    min_lots_oi: int = 50
    entry_amount_col: str = DerivativesCalculatedColumns.CloseAmount
    exit_amount_profit_col: str = DerivativesCalculatedColumns.HighAmount
    exit_amount_loss_col: str = DerivativesCalculatedColumns.LowAmount
    exit_amount_close_col: str = DerivativesCalculatedColumns.CloseAmount
    profit_factor: float = 2
    loss_factor: float = 0.5
    entry_lag_days: int = 0


class DerivativesNonTradablesDiagnostic:
    class EntryOptions:
        skip_budget_check = False
        skip_expiry_date_check = False
        skip_oi_check = False

    class ExitOptions:
        skip_holding_days_check = False
        skip_oi_check = False


class OptionsTradeBuilder:
    _trade_options: DerivativeTradeOptions
    _option_type: str

    def __init__(self, trade_options, option_type: str):
        self._trade_options = trade_options
        self._option_type = option_type

    def get_last_exit_date(self, signal_date):
        return self.get_first_entry_date(signal_date) + datetime.timedelta(
            days=self._trade_options.holding_days
        )

    def get_first_entry_date(self, signal_date):
        if self._trade_options.entry_lag_days > 0:
            for_date = pd.to_datetime(signal_date)
            for i in range(0, self._trade_options.entry_lag_days):
                for_date = MarketDaysHelper.get_next_market_day(for_date)
            return pd.to_datetime(for_date)
        else:
            return pd.to_datetime(signal_date)

    def get_full_trade(self, buy_option, sell_option, index):
        return pd.DataFrame(
            {
                DerivativesBaseColumns.Identifier: buy_option[
                    DerivativesBaseColumns.Identifier
                ].values[0],
                DerivativesBaseColumns.StrikePrice: buy_option[
                    DerivativesBaseColumns.StrikePrice
                ].values[0],
                DerivativesBaseColumns.ExpiryDate: buy_option[
                    DerivativesBaseColumns.ExpiryDate
                ].values[0],
                DerivativesBaseColumns.OptionType: buy_option[
                    DerivativesBaseColumns.OptionType
                ].values[0],
                "EntryDate": buy_option[DerivativesBaseColumns.Date].values[0],
                "Debit": -1 * buy_option["Credit"].values[0],
                "ExitDate": sell_option[DerivativesBaseColumns.Date].values[0],
                "Credit": sell_option["Credit"].values[0],
                "Quantity": 1 * buy_option["Quantity"].values[0],
                "Turnover": abs(buy_option["Credit"].values[0])
                + abs(sell_option["Credit"].values[0]),
                "Net": buy_option["Credit"].values[0] + sell_option["Credit"].values[0],
            },
            index=index,
        )

    def get_entry_option(
        self,
        data,
        symbol,
        signal_date,
        index=0,
        diagnostic=DerivativesNonTradablesDiagnostic.EntryOptions(),
    ):
        hold_till_date = self.get_last_exit_date(signal_date)
        entry_option = (
            data[
                (data[DerivativesBaseColumns.Identifier] == symbol)
                & (data[DerivativesBaseColumns.OptionType] == self._option_type)
                & (
                    data[DerivativesBaseColumns.Date]
                    == self.get_first_entry_date(signal_date)
                )
                & (
                    (
                        data[self._trade_options.entry_amount_col].between(
                            self._trade_options.budget_min,
                            self._trade_options.budget_max,
                        )
                    )
                    | (diagnostic.skip_budget_check == True)
                )
                & (
                    (
                        pd.to_datetime(data[DerivativesBaseColumns.ExpiryDate])
                        >= hold_till_date
                    )
                    | (diagnostic.skip_expiry_date_check == True)
                )
                & (
                    (
                        data[DerivativesBaseColumns.OpenInterest]
                        >= (
                            data[DerivativesCalculatedColumns.LotSize]
                            * self._trade_options.min_lots_oi
                        )
                    )
                    | (diagnostic.skip_oi_check == True)
                )
            ]
            .sort_values(DerivativesBaseColumns.StrikePrice, ascending=False)
            .head(1)
        )

        if len(entry_option) > 0:
            strike_price = entry_option[DerivativesBaseColumns.StrikePrice].values[0]
            expiry_date = entry_option[DerivativesBaseColumns.ExpiryDate].values[0]
            buy_amount = entry_option[self._trade_options.entry_amount_col].values[0]
            lotsize = entry_option[DerivativesCalculatedColumns.LotSize].values[0]
            entry_date = entry_option[DerivativesBaseColumns.Date].values[0]

            buy = pd.DataFrame(
                {
                    DerivativesBaseColumns.Identifier: symbol,
                    DerivativesBaseColumns.StrikePrice: strike_price,
                    DerivativesBaseColumns.ExpiryDate: expiry_date,
                    DerivativesBaseColumns.OptionType: self._option_type,
                    DerivativesBaseColumns.Date: entry_date,
                    "Credit": -1 * buy_amount,
                    "Quantity": 1 * lotsize,
                    "Type": "Buy",
                },
                index=[(index * 2)],
            )
            return buy
        else:
            return []

    def get_exit_option_for_buy_option(
        self,
        data,
        buy_option,
        index=0,
        diagnostic=DerivativesNonTradablesDiagnostic.ExitOptions(),
    ):
        return self.get_exit_option(
            data,
            symbol=buy_option[DerivativesBaseColumns.Identifier].values[0],
            expiry_date=buy_option[DerivativesBaseColumns.ExpiryDate].values[0],
            trade_date=buy_option[DerivativesBaseColumns.Date].values[0],
            strike_price=buy_option[DerivativesBaseColumns.StrikePrice].values[0],
            lotsize=buy_option["Quantity"].values[0],
            buy_amount=abs(buy_option["Credit"].values[0]),
            index=index,
            diagnostic=diagnostic,
        )

    def get_exit_option(
        self,
        data,
        symbol,
        expiry_date,
        trade_date,
        strike_price,
        lotsize,
        buy_amount,
        index=0,
        diagnostic=DerivativesNonTradablesDiagnostic.ExitOptions(),
    ):
        hold_till_date = self.get_last_exit_date(trade_date)
        exit_options = data[
            (data[DerivativesBaseColumns.Identifier] == symbol)
            & (data[DerivativesBaseColumns.ExpiryDate] == expiry_date)
            & (data[DerivativesBaseColumns.StrikePrice] == strike_price)
            & (data[DerivativesBaseColumns.OptionType] == self._option_type)
            & (
                (
                    data[DerivativesBaseColumns.Date].between(
                        pd.to_datetime(trade_date) + datetime.timedelta(days=1),
                        hold_till_date,
                    )
                )
                | (diagnostic.skip_holding_days_check == True)
            )
            & (
                (
                    data[DerivativesBaseColumns.OpenInterest]
                    >= (
                        data[DerivativesCalculatedColumns.LotSize]
                        * self._trade_options.min_lots_oi
                    )
                )
                | (diagnostic.skip_oi_check == True)
            )
        ].sort_values(DerivativesBaseColumns.Date)

        if len(exit_options) > 0:
            exit_option = exit_options[
                (
                    (
                        exit_options[self._trade_options.exit_amount_profit_col]
                        >= (buy_amount * self._trade_options.profit_factor)
                    )
                    | (
                        exit_options[self._trade_options.exit_amount_loss_col]
                        <= (buy_amount * self._trade_options.loss_factor)
                    )
                )
            ]

            sell_amount = 0

            if len(exit_option) <= 0:
                exit_option = exit_options.sort_values(
                    DerivativesBaseColumns.Date, ascending=False
                )[0:1]
                sell_amount = exit_option[
                    self._trade_options.exit_amount_close_col
                ].values[0]
            else:
                high_amount = exit_option[
                    self._trade_options.exit_amount_profit_col
                ].values[0]
                low_amount = exit_option[
                    self._trade_options.exit_amount_loss_col
                ].values[0]

                if high_amount >= (buy_amount * self._trade_options.profit_factor):
                    sell_amount = buy_amount * self._trade_options.profit_factor
                else:
                    sell_amount = min(
                        buy_amount * self._trade_options.loss_factor, high_amount
                    )

            # if (high_amount >= (buy_amount * self._trade_options.profit_factor)):
            #    sell_amount = (buy_amount * self._trade_options.profit_factor)
            # else:
            #    sell_amount = (buy_amount * self._trade_options.loss_factor)

            sell = pd.DataFrame(
                {
                    DerivativesBaseColumns.Identifier: symbol,
                    DerivativesBaseColumns.StrikePrice: strike_price,
                    DerivativesBaseColumns.ExpiryDate: expiry_date,
                    DerivativesBaseColumns.OptionType: self._option_type,
                    DerivativesBaseColumns.Date: pd.to_datetime(
                        exit_option[DerivativesBaseColumns.Date].values[0]
                    ),
                    "Credit": sell_amount,
                    "Quantity": -1 * lotsize,
                    "Type": "Sell",
                },
                index=[(index * 2) + 1],
            )
            return sell
        else:
            return []


class BearOptionsTradeBuilder(OptionsTradeBuilder):
    def __init__(self, trade_options):
        super().__init__(trade_options, "PE")


class BullOptionsTradeBuilder(OptionsTradeBuilder):
    def __init__(self, trade_options):
        super().__init__(trade_options, "CE")


class DerivatesMultiTradeOptions:
    avoid_overlap_trades: bool = True


class OptionsMultiTradesBuilder:
    def __init__(self, trade_options: DerivatesMultiTradeOptions):
        self.trade_options = trade_options

    def get_trades(
        self, symbol_signals, options_data, trade_builder: OptionsTradeBuilder
    ):
        trades: pd.DataFrame = pd.DataFrame()
        full_trades: pd.DataFrame = pd.DataFrame()
        non_tradables: pd.DataFrame() = pd.DataFrame()

        symbols = symbol_signals[DerivativesBaseColumns.Identifier].unique()
        for symbol in symbols:
            symbol_options_data = options_data[
                options_data[DerivativesBaseColumns.Identifier] == symbol
            ]
            for index, symbol_signal in symbol_signals[
                symbol_signals[DerivativesBaseColumns.Identifier] == symbol
            ].iterrows():
                signal_date = pd.to_datetime(symbol_signal[DerivativesBaseColumns.Date])
                if len(trades) > 0:
                    symbol_outstanding_trade_quantity = trades[
                        (trades[DerivativesBaseColumns.Identifier] == symbol)
                        & (trades[DerivativesBaseColumns.Date] <= signal_date)
                    ]["Quantity"].sum()
                else:
                    symbol_outstanding_trade_quantity = 0

                if (
                    symbol_outstanding_trade_quantity == 0
                    or self.trade_options.avoid_overlap_trades == False
                ):
                    symbol_options_data_for_dates = symbol_options_data[
                        symbol_options_data[DerivativesBaseColumns.Date].between(
                            signal_date, trade_builder.get_last_exit_date(signal_date)
                        )
                    ]

                    buy_option = trade_builder.get_entry_option(
                        symbol_options_data_for_dates, symbol, signal_date, index * 2
                    )

                    if len(buy_option) > 0:
                        sell_option = trade_builder.get_exit_option_for_buy_option(
                            symbol_options_data_for_dates, buy_option, (index * 2) + 1
                        )

                        if len(sell_option) > 0:
                            trades = pd.concat(
                                [trades, buy_option, sell_option], ignore_index=True
                            )
                            full_trades = pd.concat(
                                [
                                    full_trades,
                                    trade_builder.get_full_trade(
                                        buy_option, sell_option, index=[(index * 2)]
                                    ),
                                ],
                                ignore_index=True,
                            )
                        else:
                            non_tradables = pd.concat(
                                [
                                    non_tradables,
                                    pd.DataFrame(
                                        {
                                            DerivativesBaseColumns.Identifier: symbol,
                                            DerivativesBaseColumns.Date: pd.to_datetime(
                                                signal_date
                                            ),
                                            DerivativesBaseColumns.OptionType: trade_builder._option_type,
                                            "reason": "No Exit",
                                        },
                                        index=[(index * 2)],
                                    ),
                                ]
                            )
                    else:
                        non_tradables = pd.concat(
                            [
                                non_tradables,
                                pd.DataFrame(
                                    {
                                        DerivativesBaseColumns.Identifier: symbol,
                                        DerivativesBaseColumns.Date: pd.to_datetime(
                                            signal_date
                                        ),
                                        DerivativesBaseColumns.OptionType: trade_builder._option_type,
                                        "reason": "No Entry",
                                    },
                                    index=[(index * 2) + 1],
                                ),
                            ]
                        )

        result = TradesResult(trades, full_trades, non_tradables)
        result.prepare_summary()

        return result


class StrategyAnalyser:
    multi_trade_options: dict = {}
    trade_options: dict = {}
    results: TradesResultComposite
    output_folder: str = ""

    def __init__(self, output_sub_folder: str):
        self.results = TradesResultComposite()
        self.multi_trade_options = {}
        self.trade_options = {}
        self.output_folder = "../output/" + output_sub_folder

    def set_multi_trade_options(self, options: dict):
        self.multi_trade_options = options
        return self

    def set_trade_options(self, options: dict):
        self.trade_options = options
        return self

    def iterator(self, callback):
        for key1 in self.multi_trade_options:
            for key2 in self.trade_options:
                callback(
                    key1, key2, self.multi_trade_options[key1], self.trade_options[key2]
                )

    def load_from_files(self):
        for key1 in self.multi_trade_options:
            for key2 in self.trade_options:
                result = BiDirectionalTradesResult.load_from_file(
                    f"{self.output_folder}/{key1}-{key2}.xlsx"
                )
                self.results.add_result(result, f"{key1}-{key2}")
        return self.results

    def analyze(
        self,
        options_data: pd.DataFrame,
        sell_signal_data_file: str = "sell_signal_symbols.csv",
        buy_signal_data_file: str = "buy_signal_symbols.csv",
    ):
        sell_signal_data = pd.read_csv(f"{self.output_folder}/{sell_signal_data_file}")
        buy_signal_data = pd.read_csv(f"{self.output_folder}/{buy_signal_data_file}")

        sell_signal_data[DerivativesBaseColumns.Date] = pd.to_datetime(
            sell_signal_data[DerivativesBaseColumns.Date]
        )
        buy_signal_data[DerivativesBaseColumns.Date] = pd.to_datetime(
            buy_signal_data[DerivativesBaseColumns.Date]
        )

        sell_signal_symbols = sell_signal_data[
            DerivativesBaseColumns.Identifier
        ].unique()
        buy_signal_symbols = buy_signal_data[DerivativesBaseColumns.Identifier].unique()

        all_symbols = np.unique(
            np.concatenate((sell_signal_symbols, buy_signal_symbols))
        )

        options_data = options_data[
            options_data[DerivativesBaseColumns.Identifier].isin(all_symbols)
        ]

        self.results = TradesResultComposite()
        for key1 in self.multi_trade_options:
            multi_trade_builder = OptionsMultiTradesBuilder(
                self.multi_trade_options[key1]
            )
            for key2 in self.trade_options:
                bear_trade_builder = BearOptionsTradeBuilder(self.trade_options[key2])
                bull_trade_builder = BullOptionsTradeBuilder(self.trade_options[key2])
                bear_trades = multi_trade_builder.get_trades(
                    sell_signal_data, options_data, bear_trade_builder
                )
                bull_trades = multi_trade_builder.get_trades(
                    buy_signal_data, options_data, bull_trade_builder
                )

                result = BiDirectionalTradesResult(bear_trades, bull_trades)
                result.save_result(f"{self.output_folder}/{key1}-{key2}.xlsx")
                self.results.add_result(result, f"{key1}-{key2}")
        # self.results.save_result(f'{self.output_folder}/final.xlsx')
        return self.results



from markets_insights.core.column_definition import (
    DerivativesBaseColumns,
    DerivativesCalculatedColumns,
)
from markets_insights.core.core import Instrumentation
from markets_insights.dataprocess.data_processor import CalculationWorker
import pandas as pd


class DerivativesPriceCalculationWorker(CalculationWorker):
    @Instrumentation.trace(name="DerivativesPriceCalculationWorker")
    def add_calculated_columns(self, data):
        if DerivativesBaseColumns.PreviousClose not in data.columns.to_list():
            data[DerivativesBaseColumns.PreviousClose] = data.groupby(
                [
                    DerivativesBaseColumns.Identifier,
                    DerivativesBaseColumns.OptionType,
                    DerivativesBaseColumns.ExpiryDate,
                    DerivativesBaseColumns.StrikePrice,
                ]
            )[DerivativesBaseColumns.Close].transform(lambda x: x.shift(1))

        data[DerivativesCalculatedColumns.CloseToPrevCloseChangePerc] = (
            data[DerivativesBaseColumns.Close]
            / data[DerivativesBaseColumns.PreviousClose]
            - 1
        ) * 100
        data[DerivativesCalculatedColumns.OpenToPrevCloseChangePerc] = (
            data[DerivativesBaseColumns.Open]
            / data[DerivativesBaseColumns.PreviousClose]
            - 1
        ) * 100

        data[DerivativesCalculatedColumns.PreviousCloseAmount] = (
            data[DerivativesBaseColumns.PreviousClose]
            * data[DerivativesCalculatedColumns.LotSize]
        )
        data[DerivativesCalculatedColumns.OpenAmount] = (
            data[DerivativesBaseColumns.Open]
            * data[DerivativesCalculatedColumns.LotSize]
        )
        data[DerivativesCalculatedColumns.CloseAmount] = (
            data[DerivativesBaseColumns.Close]
            * data[DerivativesCalculatedColumns.LotSize]
        )
        data[DerivativesCalculatedColumns.HighAmount] = (
            data[DerivativesBaseColumns.High]
            * data[DerivativesCalculatedColumns.LotSize]
        )
        data[DerivativesCalculatedColumns.LowAmount] = (
            data[DerivativesBaseColumns.Low]
            * data[DerivativesCalculatedColumns.LotSize]
        )

        data[DerivativesCalculatedColumns.AmountDiffOpenToPrevClose] = (
            data[DerivativesCalculatedColumns.OpenAmount]
            - data[DerivativesCalculatedColumns.PreviousCloseAmount]
        )
        data[DerivativesCalculatedColumns.AmountDiffCloseToPrevClose] = (
            data[DerivativesCalculatedColumns.CloseAmount]
            - data[DerivativesCalculatedColumns.PreviousCloseAmount]
        )


class DerivativesLotSizeCalculationWorker(CalculationWorker):
    @Instrumentation.trace(name="DerivativesLotSizeCalculationWorker")
    def add_calculated_columns(self, data):
        data[DerivativesCalculatedColumns.LotSize] = data.groupby(
            [DerivativesBaseColumns.Identifier, DerivativesBaseColumns.ExpiryDate]
        )[DerivativesBaseColumns.OpenInterest].transform(lambda x: x.min())


class DerivativesLotSizeCalculationWorker_Obsolete(CalculationWorker):
    @Instrumentation.trace(name="DerivativesLotSizeCalculationWorker")
    def add_calculated_columns(self, data):
        lotsize_data = pd.read_csv("../manual_data/nse_fo_lotsize.csv")
        cur_expiry_col = lotsize_data.columns[2]
        data = pd.merge(
            data,
            lotsize_data[["SYMBOL", cur_expiry_col]],
            how="left",
            left_on=DerivativesBaseColumns.Identifier,
            right_on="SYMBOL",
        )
        data.rename(
            columns={cur_expiry_col: DerivativesCalculatedColumns.LotSize}, inplace=True
        )
        return data



from markets_insights.core.column_definition import (
    BaseColumns,
    CalculatedColumns,
    DerivativesBaseColumns,
    DerivativesCalculatedColumns,
)
from markets_insights.core.core import Instrumentation
from markets_insights.calculations.base import CalculationWorker
import pandas as pd
from datetime import datetime
from dateutil.relativedelta import relativedelta
from markets_insights.datareader.data_reader import NseDerivatiesOldReader
from markets_insights.core.core import MarketDaysHelper


class IsInDerivativesFlagCalculationWorker(CalculationWorker):
    @Instrumentation.trace(name="IsInDerivativesFlagCalculationWorker")
    def add_calculated_columns(self, data):
        from_date = data[BaseColumns.Date].min()
        to_date = data[BaseColumns.Date].max()

        all_symbols = pd.DataFrame()
        current_date = datetime(from_date.year, from_date.month, 1)
        derivatives_reader = NseDerivatiesOldReader()

        while current_date <= to_date:
            derivatives_symbols = derivatives_reader.read(
                MarketDaysHelper.get_this_or_next_market_day(current_date)
            )[[DerivativesBaseColumns.Identifier, DerivativesBaseColumns.Date]]
            derivatives_symbols[
                DerivativesCalculatedColumns.MonthNo
            ] = current_date.month
            derivatives_symbols[DerivativesCalculatedColumns.Year] = current_date.year
            derivatives_symbols.drop_duplicates(inplace=True)
            all_symbols = pd.concat(
                [all_symbols, derivatives_symbols], ignore_index=True
            )
            current_date = current_date + relativedelta(months=+1)

            all_symbols.drop_duplicates(inplace=True)
            all_symbols.drop(columns=[BaseColumns.Date], inplace=True)
            all_symbols[CalculatedColumns.IsInDerivatives] = True

        data = pd.merge(
            data,
            all_symbols,
            how="left",
            on=[
                BaseColumns.Identifier,
                CalculatedColumns.Year,
                CalculatedColumns.MonthNo,
            ],
        )
        return data
